@@ -7,6 +7,7 @@
 -- | Types for possible backends to run client-side `Request` queries
 module Servant.Client.Core.RunClient (
     RunClient (..),
+    runRequest,
     RunStreamingClient (..),
     ClientF (..),
     ) where
@@ -14,6 +15,8 @@ module Servant.Client.Core.RunClient (
 import           Prelude ()
 import           Prelude.Compat
 
+import           Network.HTTP.Types.Status
+                 (Status)
 import           Control.Monad.Free
                  (Free (..), liftF)
 
@@ -22,10 +25,15 @@ import           Servant.Client.Core.Request
 import           Servant.Client.Core.Response
 
 class Monad m => RunClient m where
-  -- | How to make a request.
-  runRequest :: Request -> m Response
+  -- | How to make a request, with an optional list of status codes to not throw exceptions
+  -- for (default: [200..299]).
+  runRequestAcceptStatus :: Maybe [Status] -> Request -> m Response
   throwClientError :: ClientError -> m a
 
+-- | How to make a request.
+runRequest :: RunClient m => Request -> m Response
+runRequest = runRequestAcceptStatus Nothing
+
 class RunClient m =>  RunStreamingClient m where
     withStreamingRequest :: Request -> (StreamingResponse -> IO a) ->  m a
 
@@ -41,6 +49,7 @@ data ClientF a
     | Throw ClientError
   deriving (Functor)
 
+-- TODO: honour the accept-status argument.
 instance ClientF ~ f => RunClient (Free f) where
-    runRequest req  = liftF (RunRequest req id)
+    runRequestAcceptStatus _ req  = liftF (RunRequest req id)
     throwClientError = liftF . Throw
